from sys import exit

import click

import rcds
import rcds.challenge.docker
from rcds.util import SUPPORTED_EXTENSIONS, find_files


@click.command()
def deploy() -> None:
    try:
        project_config = find_files(["rcds"], SUPPORTED_EXTENSIONS, recurse=True)[
            "rcds"
        ].parent
    except KeyError:
        click.echo("Could not find project root!")
        exit(1)
    click.echo(f"Loading project at {project_config}")
    project = rcds.Project(project_config)
    click.echo("Initializing backends")
    project.load_backends()
    click.echo("Loading challenges")
    project.load_all_challenges()
    for challenge in project.challenges.values():
        cm = rcds.challenge.docker.ContainerManager(challenge)
        for container_name, container in cm.containers.items():
            click.echo(f"{challenge.config['id']}: checking container {container_name}")
            if not container.is_built():
                click.echo(
                    f"{challenge.config['id']}: building container {container_name}"
                    f" ({container.get_full_tag()})"
                )
                container.build()
        challenge.create_transaction().commit()
    if project.container_backend is not None:
        click.echo("Commiting container backend")
        project.container_backend.commit()
    else:
        click.echo("WARN: no container backend, skipping...")
    if project.scoreboard_backend is not None:
        click.echo("Commiting scoreboard backend")
        project.scoreboard_backend.commit()
    else:
        click.echo("WARN: no scoreboard backend, skipping...")



import re
from typing import Any, Callable, Dict, Iterable, List, Set

from kubernetes import client  # type: ignore

AnyManifest = Dict[str, Any]


# namespaced manifests only - namespaces are handled separately
#MANIFEST_KINDS = ["Secret", "Deployment", "Service", "Ingress", "NetworkPolicy"]
MANIFEST_KINDS = ["Deployment", "Service", "Ingress", "NetworkPolicy"]
KIND_TO_API_VERISON = {
    "Deployment": "apps/v1",
    "Service": "v1",
    "Ingress": "networking.k8s.io/v1",
    "NetworkPolicy": "networking.k8s.io/v1",
    "Secret": "v1",
}


camel_case_to_snake_case_re = re.compile(r"(?=[A-Z])")


def kind_to_api_method_postfix(kind: str) -> str:
    return "_namespaced" + camel_case_to_snake_case_re.sub("_", kind).lower()


def get_api_method_for_kind(api_client: Any, method: str, kind: str) -> Callable:
    return getattr(api_client, method + kind_to_api_method_postfix(kind))


def labels_to_label_selector(labels: Dict[str, str]) -> str:
    selector = ""
    for k, v in labels.items():
        selector += f"{k}={v},"
    return selector[:-1]


def sync_manifests(all_manifests: Iterable[Dict[str, Any]]):
    v1 = client.CoreV1Api()
    appsv1 = client.AppsV1Api()
    networkingv1 = client.NetworkingV1Api()
    #networkingv1beta1 = client.NetworkingV1beta1Api()

    api_version_to_client = {
        "v1": v1,
        "apps/v1": appsv1,
        "networking.k8s.io/v1": networkingv1,
        #"networking.k8s.io/v1beta1": networkingv1beta1,
    }

    manifests_by_namespace_kind: Dict[str, Dict[str, List[Dict[str, Any]]]] = dict()
    namespaces: List[Dict[str, Any]] = []

    for manifest in all_manifests:
        kind = manifest["kind"]
        if kind == "Namespace":
            namespaces.append(manifest)
        else:
            namespace = manifest["metadata"]["namespace"]
            manifests_by_namespace_kind.setdefault(namespace, dict())
            manifests_by_namespace_kind[namespace].setdefault(kind, [])
            manifests_by_namespace_kind[namespace][kind].append(manifest)

    server_namespaces_names: Set[str] = set(
        map(
            lambda ns: ns.metadata.name,
            v1.list_namespace(label_selector="app.kubernetes.io/managed-by=rcds").items,
        )
    )

    for namespace_manifest in namespaces:
        namespace = namespace_manifest["metadata"]["name"]

        try:
            server_namespaces_names.remove(namespace)
            # the namespace already exists; patch it
            print(f"PATCH Namespace {namespace}")
            v1.patch_namespace(namespace, namespace_manifest)
        except KeyError:
            # the namespace doesn't exist; create it
            print(f"CREATE Namespace {namespace}")
            v1.create_namespace(namespace_manifest)

        # TODO: Potentially decouple this from the namespace's labels?
        # Common labels for rCDS manifests in this namespace
        ns_labels: Dict[str, str] = namespace_manifest["metadata"]["labels"]
        ns_labels.pop("name")

        # Process all manifest kinds we know about in this namespace
        for kind in MANIFEST_KINDS:
            manifests = manifests_by_namespace_kind[namespace].get(kind, [])
            server_manifest_names: Set[str] = set(
                map(
                    lambda m: m.metadata.name,
                    get_api_method_for_kind(
                        api_version_to_client[KIND_TO_API_VERISON[kind]], "list", kind
                    )(
                        namespace, label_selector=labels_to_label_selector(ns_labels)
                    ).items,
                )
            )
            for manifest in manifests:
                manifest_name = manifest["metadata"]["name"]
                try:
                    server_manifest_names.remove(manifest_name)
                    # the manifest already exists; patch it
                    print(f"PATCH {kind} {namespace}/{manifest_name}")
                    try:
                        get_api_method_for_kind(
                            api_version_to_client[manifest["apiVersion"]], "patch", kind
                        )(manifest_name, namespace, manifest)
                    except client.rest.ApiException:
                        # Conflict of some sort - let's just delete and recreate it
                        print(f"DELETE {kind} {namespace}/{manifest_name}")
                        get_api_method_for_kind(
                            api_version_to_client[manifest["apiVersion"]],
                            "delete",
                            kind,
                        )(manifest_name, namespace)
                        print(f"CREATE {kind} {namespace}/{manifest_name}")
                        get_api_method_for_kind(
                            api_version_to_client[manifest["apiVersion"]],
                            "create",
                            kind,
                        )(namespace, manifest)
                except KeyError:
                    # the manifest doesn't exist; create it
                    print(f"CREATE {kind} {namespace}/{manifest_name}")
                    get_api_method_for_kind(
                        api_version_to_client[manifest["apiVersion"]], "create", kind
                    )(namespace, manifest)
            for manifest_name in server_manifest_names:
                print(f"DELETE {kind} {namespace}/{manifest_name}")
                get_api_method_for_kind(
                    api_version_to_client[KIND_TO_API_VERISON[kind]], "delete", kind
                )(manifest_name, namespace)

    for namespace_name in server_namespaces_names:
        print(f"DELETE Namespace {namespace_name}")
        v1.delete_namespace(namespace_name)



import os
import itertools
from pathlib import Path
from typing import Any, Dict, List

import yaml
from jinja2 import Environment, Template
from kubernetes import config  # type: ignore

import rcds
import rcds.backend
from rcds.util import load_any
from rcds.util.jsonschema import DefaultValidatingDraft7Validator

from .jinja import jinja_env
from .manifests import AnyManifest, sync_manifests

options_schema_validator = DefaultValidatingDraft7Validator(
    schema=load_any(Path(__file__).parent / "options.schema.yaml")
)


class ContainerBackend(rcds.backend.BackendContainerRuntime):
    _project: rcds.Project
    _options: Dict[str, Any]
    _namespace_template: Template
    _jinja_env: Environment

    def __init__(self, project: rcds.Project, options: Dict[str, Any]):
        self._project = project
        self._options = options

        for option_key in ["dockerConfigJson"]:
            env_key = f"RCDS_K8S_{option_key.upper()}"
            env = os.environ.get(env_key)
            if env is not None:
                self._options[option_key] = env

        crt = os.environ.get("RCDS_K8S_TLS_CRT")
        key = os.environ.get("RCDS_K8S_TLS_KEY")
        if crt and key:
            cert = {
                "crt": crt,
                "key": key,
            }
            self._options["tlsCertificate"] = cert

        # FIXME: validate options better
        if not options_schema_validator.is_valid(self._options):
            raise ValueError("Invalid options")

        self._namespace_template = Template(self._options["namespaceTemplate"])
        self._jinja_env = jinja_env.overlay()
        self._jinja_env.globals["options"] = self._options

        config.load_kube_config(context=self._options.get("kubeContext", None))

    def commit(self) -> bool:
        deployed_challs = filter(
            lambda c: c.config["deployed"], self._project.challenges.values()
        )
        # TODO: auto assignment of expose params
        manifests = list(
            itertools.chain.from_iterable(
                map(
                    lambda chall: self.gen_manifests_for_challenge(chall),
                    deployed_challs,
                )
            )
        )
        sync_manifests(manifests)
        return True

    def get_namespace_for_challenge(self, challenge: rcds.Challenge) -> str:
        return self._namespace_template.render({"challenge": challenge.config})

    def gen_manifests_for_challenge(
        self, challenge: rcds.Challenge
    ) -> List[AnyManifest]:
        if "containers" not in challenge.config:
            return []

        manifests: List[Dict[str, Any]] = []

        def render_and_append(env: Environment, template: str) -> None:
            nonlocal manifests
            manifest = env.get_template(template).render().strip()
            manifests += filter(lambda x: x is not None, yaml.safe_load_all(manifest))

        challenge_env: Environment = self._jinja_env.overlay()
        challenge_env.globals["challenge"] = challenge.config
        challenge_env.globals["namespace"] = self.get_namespace_for_challenge(challenge)

        render_and_append(challenge_env, "namespace.yaml")
        render_and_append(challenge_env, "network-policy.yaml")

        if self._options.get("dockerConfigJson"):
            render_and_append(challenge_env, "secret.yaml")

        for container_name, container_config in challenge.config["containers"].items():
            expose_config = challenge.config.get("expose", dict()).get(
                container_name, None
            )

            if expose_config is not None:
                for expose_port in expose_config:
                    if "http" in expose_port:
                        if isinstance(expose_port["http"], str):
                            expose_port["http"] += "." + self._options["domain"]
                        else:
                            assert isinstance(expose_port["http"], dict)
                            if "raw" in expose_port["http"]:
                                expose_port["http"] = expose_port["http"]["raw"]
                    if "tcp" in expose_port:
                        expose_port["host"] = self._options["domain"]

            container_env: Environment = challenge_env.overlay()
            container_env.globals["container"] = {
                "name": container_name,
                "config": container_config,
            }
            if expose_config is not None:
                container_env.globals["container"]["expose"] = expose_config

            render_and_append(container_env, "deployment.yaml")
            render_and_append(container_env, "service.yaml")
            render_and_append(container_env, "ingress.yaml")

        return manifests


class BackendsInfo(rcds.backend.BackendsInfo):
    HAS_CONTAINER_RUNTIME = True

    def get_container_runtime(
        self, project: rcds.Project, options: Dict[str, Any]
    ) -> ContainerBackend:
        return ContainerBackend(project, options)


def get_info() -> BackendsInfo:
    return BackendsInfo()



import base64
import collections.abc
import hashlib
import json
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Any, Dict, Iterable, Iterator, Type, Union, cast

import docker  # type: ignore
import pathspec  # type: ignore

if TYPE_CHECKING:
    from ..project import Project
    from .challenge import Challenge


def flatten(i: Iterable[Union[str, Iterable[str]]]) -> Iterable[str]:
    for x in i:
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, str):
            yield from (y for y in x)
        else:
            yield x


def get_context_files(root: Path) -> Iterator[Path]:
    """
    Generate a list of all files in the build context of the specified Dockerfile

    :param pathlib.Path root: Path to the containing directory of the Dockerfile to
        analyze
    """
    files: Iterator[Path] = root.rglob("*")
    dockerignore = root / ".dockerignore"
    if dockerignore.exists():
        with dockerignore.open("r") as fd:
            spec = pathspec.PathSpec.from_lines(
                "gitwildmatch",
                flatten(
                    # pathspec's behavior with negated patterns is different than that
                    # of docker (and its own behavior with non-negated patterns) in that
                    # patterns ending with `/` will match files in subdirectories, but
                    # not a file with the same name, and pattens not ending in `/` will
                    # only match files, but not files in subdirectories. For example,
                    # the pattern `!/a` will exclude `a`, but not `a/b`, and the pattern
                    # `!/a/` will exclude `a/b`, but not `a`. Since docker treats these
                    # interchangeably, we automatically insert the corresponding ignore
                    # rule into the rules list if a negated pattern is detected (insert
                    # `!/a/` if `!/a` is detected and vice versa).
                    # FIXME: normalize / parse the lines better to support e.g. comments
                    (
                        (
                            [line[:-2] + "\n", line]
                            if line[-2] == "/"
                            else [line, line[:-1] + "/\n"]
                        )
                        if line[0] == "!"
                        else line
                    )
                    for line in fd
                ),
            )
        files = filter(lambda p: not spec.match_file(p.relative_to(root)), files)
    return filter(lambda p: p.is_file(), files)


def generate_sum(root: Path) -> str:
    """
    Generate a checksum of all files in the build context of the specified directory

    :param pathlib.Path root: Path to the containing directory of the Dockerfile to
        analyze
    """
    h = hashlib.sha256()
    for f in sorted(get_context_files(root), key=lambda f: str(f.relative_to(root))):
        h.update(bytes(f.relative_to(root)))
        with f.open("rb") as fd:
            for chunk in iter(lambda: fd.read(524288), b""):
                h.update(chunk)
    return h.hexdigest()


class Container:
    """
    A single container
    """

    manager: "ContainerManager"
    challenge: "Challenge"
    project: "Project"
    name: str
    config: Dict[str, Any]

    IS_BUILDABLE: bool = False

    def __init__(self, *, container_manager: "ContainerManager", name: str) -> None:
        self.manager = container_manager
        self.challenge = self.manager.challenge
        self.project = self.challenge.project
        self.name = name
        self.config = container_manager.config[self.name]

    def get_full_tag(self) -> str:
        """
        Get the full image tag (e.g. ``k8s.gcr.io/etcd:3.4.3-0``) for this container

        :returns: The image tag
        """
        return self.config["image"]

    def is_built(self) -> bool:
        """
        If the container is buildable (:const:`IS_BUILDABLE` is `True`), this method
        returns whether or not the container is already built (and up-to-date). For
        non-buildable containers, this method always returns `True`.

        :returns: Whether or not the container is built
        """
        return True

    def build(self, force: bool = False) -> None:
        """
        Build the challenge if applicable and necessary.

        For challenges that are not buildable (:const:`IS_BUILDABLE` is False), this
        method is a no-op

        :param bool force: Force a rebuild of this container even if it is up-to-date
        """
        pass


class BuildableContainer(Container):
    """
    A container that is built from source
    """

    root: Path
    dockerfile: str
    buildargs: Dict[str, str]

    IS_BUILDABLE: bool = True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        build = self.config.get("build", None)
        assert build is not None
        if isinstance(build, str):
            self.root = self.challenge.root / Path(build)
            self.dockerfile = "Dockerfile"
            self.buildargs = dict()
        elif isinstance(build, dict):
            build = cast(Dict[str, Any], build)
            self.root = self.challenge.root / Path(build["context"])
            self.dockerfile = build.get("dockerfile", "Dockerfile")
            self.buildargs = cast(Dict[str, str], build.get("args", dict()))
        self.content_hash = generate_sum(self.root)
        self.image = self.manager.get_docker_image(self)

    def _build(self) -> None:
        self.project.docker_client.images.build(
            path=str(self.root),
            tag=f"{self.image}:{self.content_hash}",
            dockerfile=self.dockerfile,
            buildargs=self.buildargs,
            pull=True,
            rm=True,
        )
        self.project.docker_client.images.push(
            self.image, tag=self.content_hash, auth_config=self.manager._auth_config
        )

    def get_full_tag(self) -> str:
        return f"{self.image}:{self.content_hash}"

    def is_built(self) -> bool:
        """
        Checks if a container built with a build context with a matching hash exists,
        either locally or remotely.

        :returns: Whether or not the image was found
        """
        try:
            self.project.docker_client.images.get_registry_data(
                self.get_full_tag(), auth_config=self.manager._auth_config
            )
            return True
        except docker.errors.NotFound:
            pass  # continue
        return False

    def build(self, force: bool = False) -> None:
        if force or not self.is_built():
            self._build()


class _AuthCfgCache:
    _cache: Dict[str, Dict[str, str]] = dict()  # class-level

    def get_auth_config(self, registry: str, api_client) -> Dict[str, str]:
        if registry not in self._cache:
            header = docker.auth.get_config_header(api_client, registry)
            if header is not None:
                auth_config = json.loads(
                    base64.urlsafe_b64decode(header).decode("ascii")
                )
            else:
                auth_config = None
            self._cache[registry] = auth_config
        return self._cache[registry]


_auth_cfg_cache = _AuthCfgCache()


class ContainerManager:
    """
    Object managing all containers defined by a given :class:`rcds.Challenge`
    """

    challenge: "Challenge"
    project: "Project"
    config: Dict[str, Dict[str, Any]]
    containers: Dict[str, Container]
    _auth_config: Dict[str, str]

    def __init__(self, challenge: "Challenge"):
        """
        :param rcds.Challenge challenge: The challenge that this ContainerManager
            belongs to
        """

        self.challenge = challenge
        self.project = self.challenge.project
        self.containers = dict()
        self.config = cast(
            Dict[str, Dict[str, Any]], self.challenge.config.get("containers", dict())
        )

        self._auth_config = self._get_auth_config()

        for name in self.config.keys():
            container_config = self.config[name]
            container_constructor: Type[Container]
            if "build" in container_config:
                container_constructor = BuildableContainer
            else:
                container_constructor = Container
            self.containers[name] = container_constructor(
                container_manager=self, name=name
            )
            container_config["image"] = self.containers[name].get_full_tag()

    def get_docker_image(self, container: Container) -> str:
        image_template = self.project.jinja_env.from_string(
            self.project.config["docker"]["image"]["template"]
        )
        template_context = {
            "challenge": self.challenge.config,
            "container": dict(container.config),
        }
        template_context["container"]["name"] = container.name
        image = image_template.render(template_context)
        # FIXME: better implementation than abusing PosixPath?
        return str(
            PurePosixPath(self.project.config["docker"]["image"]["prefix"]) / image
        )

    def _get_auth_config(self) -> Dict[str, str]:
        registry, _ = docker.auth.resolve_repository_name(
            self.project.config["docker"]["image"]["prefix"]
        )
        return _auth_cfg_cache.get_auth_config(registry, self.project.docker_client.api)



import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, cast

from ..util import SUPPORTED_EXTENSIONS, deep_merge, find_files
from .config import ConfigLoader

if TYPE_CHECKING:
    import rcds

    from ..project import Project
    from ..project.assets import AssetManagerContext, AssetManagerTransaction


def _strip_scheme(url: str) -> str:
    return re.sub(r".*?://", "", url)


class ChallengeLoader:
    """
    Class for loading a :class:`Challenge` within the context of a
    :class:`rcds.Project`
    """

    project: "Project"
    _config_loader: ConfigLoader

    def __init__(self, project: "rcds.Project"):
        self.project = project
        self._config_loader = ConfigLoader(self.project)

    def load(self, root: Path):
        """
        Load a challenge by path

        The challenge must be within the project associated with this loader.

        :param pathlib.Path root: Path to challenge root
        """
        try:
            cfg_file = find_files(
                ["challenge"], SUPPORTED_EXTENSIONS, path=root, recurse=False
            )["challenge"]
        except KeyError:
            raise ValueError(f"No config file found at '{root}'")
        config = self._config_loader.load_config(cfg_file)
        return Challenge(self.project, root, config)


class Challenge:
    """
    A challenge within a given :class:`rcds.Project`

    This class is not meant to be constructed directly, use a :class:`ChallengeLoader`
    to load a challenge.
    """

    project: "Project"
    root: Path
    config: Dict[str, Any]
    context: Dict[str, Any]  # overrides to Jinja context
    _asset_manager_context: "AssetManagerContext"
    _asset_sources: List[Callable[["AssetManagerTransaction"], None]]

    def __init__(self, project: "Project", root: Path, config: dict):
        self.project = project
        self.root = root
        self.config = config
        self.context = dict()
        self._asset_manager_context = self.project.asset_manager.create_context(
            self.config["id"]
        )
        self._asset_sources = []

        self.register_asset_source(self._add_static_assets)

    def _add_static_assets(self, transaction: "AssetManagerTransaction") -> None:
        if "provide" not in self.config:
            return
        for provide in self.config["provide"]:
            if isinstance(provide, str):
                path = self.root / Path(provide)
                name = path.name
            else:
                path = self.root / Path(provide["file"])
                name = provide["as"]
            transaction.add_file(name, path)

    def register_asset_source(
        self, do_add: Callable[["AssetManagerTransaction"], None]
    ) -> None:
        """
        Register a function to add assets to the transaction for this challenge.
        """
        self._asset_sources.append(do_add)

    def create_transaction(self) -> "AssetManagerTransaction":
        """
        Get a transaction to update this challenge's assets
        """
        transaction = self._asset_manager_context.transaction()
        for do_add in self._asset_sources:
            do_add(transaction)
        return transaction

    def get_asset_manager_context(self) -> "AssetManagerContext":
        return self._asset_manager_context

    def get_relative_path(self) -> Path:
        """
        Utiity function to get this challenge's path relative to the project root
        """
        return self.root.relative_to(self.project.root)

    def get_context_shortcuts(self) -> Dict[str, Any]:
        shortcuts: Dict[str, Any] = dict()

        if (
            "expose" in self.config
            and len(self.config["expose"]) == 1
            and len(next(iter(cast(Dict[str, list], self.config["expose"]).values())))
            == 1
        ):
            # One container exposed; we can define expose shortcuts
            expose_cfg = cast(
                Dict[str, Any], next(iter(self.config["expose"].values()))[0]
            )
            shortcuts["host"] = expose_cfg.get("http", expose_cfg.get("host", None))
            has_url = False
            if "tcp" in expose_cfg:
                shortcuts["port"] = expose_cfg["tcp"]
                shortcuts["nc"] = f"nc {shortcuts['host']} {shortcuts['port']}"
                shortcuts["url"] = f"http://{shortcuts['host']}:{shortcuts['port']}"
                has_url = True
            if "http" in expose_cfg:
                shortcuts["url"] = f"https://{shortcuts['host']}"
                has_url = True
            if has_url:
                shortcuts[
                    "link"
                ] = f"[{_strip_scheme(shortcuts['url'])}]({shortcuts['url']})"

        return shortcuts

    def render_description(self) -> str:
        """
        Render the challenge's description template to a string
        """

        return self.project.jinja_env.from_string(self.config["description"]).render(
            deep_merge(
                dict(),
                {"challenge": self.config},
                self.get_context_shortcuts(),
                self.context,
            )
        )



import base64
from textwrap import dedent
from typing import Any, Dict, Optional

import yaml
from jinja2 import Environment, PackageLoader, filters

jinja_env = Environment(
    loader=PackageLoader("rcds.backends.k8s", "templates"),
    autoescape=False,
    trim_blocks=True,
    lstrip_blocks=True,
)


def jinja_filter_indent(data: str, *args, **kwargs) -> str:
    return filters.do_indent(dedent(data), *args, **kwargs)


def jinja_filter_yaml(data: Dict[str, Any], indent: Optional[int] = None) -> str:
    output = yaml.dump(data).strip()
    if indent is not None:
        output = jinja_filter_indent(output, indent)
    return output


jinja_env.filters["indent"] = jinja_filter_indent
jinja_env.filters["yaml"] = jinja_filter_yaml
jinja_env.filters["quote"] = lambda s: repr(str(s))
jinja_env.filters["b64encode"] = lambda s: base64.b64encode(s.encode('ascii')).decode('ascii')


import io
import json
import os
import pathlib
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    ByteString,
    Callable,
    Dict,
    Iterable,
    Set,
    Tuple,
    Union,
    cast,
)
from warnings import warn

if TYPE_CHECKING:
    import rcds

    from .project import Project


File = Union[BinaryIO, Path, bytes]
"""
Something that the asset manager can interpret as a file (contents only)

Valid types:

- A :class:`pathlib.Path` object referring to a file that already exists on-disk

- Any :class:`typing.BinaryIO` object that is seekable

- A :class:`typing.ByteString` object containing the contents of the file (internally
  this is converted to a :class:`io.BytesIO`)
"""


def _is_valid_name(name: str):
    return (
        len(pathlib.PurePosixPath(name).parts) == 1
        and len(pathlib.PureWindowsPath(name).parts) == 1
    )


class AssetManagerTransaction:
    """
    A transaction within an :class:`AssetManagerContext`

    This class manages declarative transactional updates to a context, allowing you to
    declare the files that should exist in the context, the last time that file was
    modified, and a callable to run to get the file, should it be out-of-date in the
    cache. The transaction starts in a blank state; without adding anything by calling
    :meth:`add`, :meth:`commit` will clear the context. No actions are performed until
    :meth:`commit` is called.

    This classs is not meant to be constructed directly, use
    :meth:`AssetManagerContext.transaction`
    """

    _asset_manager_context: "AssetManagerContext"
    _is_active: bool

    @dataclass
    class _FileEntry:
        """
        :meta private:
        """

        mtime: float

        # Callable is wrapped in a tuple because otherwise, mypy thinks the field is a
        # class method (related to python/mypy#708)
        get_contents: Tuple[Callable[[], File]]

    _files: Dict[str, _FileEntry]

    def __init__(self, asset_manager_context: "AssetManagerContext"):
        """
        :meta private:
        """
        self._asset_manager_context = asset_manager_context
        self._is_active = True
        self._files = dict()

    def add(
        self, name: str, mtime: float, contents: Union[File, Callable[[], File]]
    ) -> None:
        """
        Add a file to the context

        :param str name: The name of the asset to add
        :param float mtime: The time the asset to add was modified
            (:attr:`os.stat_result.st_mtime`)
        :param contents: The contents of the file - this can either be the contents
            directly as a :const:`File`, or a thunk function that, when calls, returns
            the contents
        :type contents: :const:`File` or :obj:`Callable[[], File]`
        :raises RuntimeError: if the transaction has already been committed
        :raises ValueError: if the asset name is not valid
        """
        if not self._is_active:
            raise RuntimeError("This transaction has already been committed")
        self._asset_manager_context._assert_valid_name(name)
        get_contents: Callable[[], File]
        if callable(contents):
            get_contents = contents
        else:

            def get_contents() -> File:
                return cast(File, contents)

        self._files[name] = self._FileEntry(mtime=mtime, get_contents=(get_contents,))

    def add_file(self, name: str, file: Path):
        """
        Add an already-existing file on disk to the context

        This wraps :meth:`add`

        :param str name: The name of the asset to add
        :param Path file: The path to the asset on disk
        """
        if not file.is_file():
            raise ValueError(f"Provided file does not exist: '{str(file)}'")
        self.add(name, file.stat().st_mtime, lambda: file)

    def _create(self, fpath: Path, fentry: _FileEntry) -> None:
        """
        Create / overwrite the asset in the cache

        :meta private:
        """
        contents = fentry.get_contents[0]()
        if isinstance(contents, Path):
            if not contents.is_file():
                raise ValueError(f"Provided file does not exist: '{str(contents)}'")
            fpath.symlink_to(contents)
        else:
            if isinstance(contents, ByteString):
                contents = io.BytesIO(contents)
            assert isinstance(contents, io.IOBase)
            with fpath.open("wb") as ofd:
                shutil.copyfileobj(contents, ofd)
        os.utime(fpath, (fentry.mtime, fentry.mtime))

    def commit(self) -> None:
        """
        Commit the transaction.

        This transaction can no longer be used after :meth:`commit` is called.
        """
        self._is_active = False
        self._asset_manager_context._is_transaction_active = False
        files_to_delete = set(self._asset_manager_context.ls())
        for name, file_entry in self._files.items():
            fpath = self._asset_manager_context._get(name)
            try:
                files_to_delete.remove(name)
            except KeyError:
                pass
            if self._asset_manager_context.exists(name):
                cache_mtime = self._asset_manager_context.get_mtime(name)
                if not file_entry.mtime > cache_mtime:
                    continue
            self._create(fpath, file_entry)
            self._asset_manager_context._add(name, force=True)
        for name in files_to_delete:
            fpath = self._asset_manager_context.get(name)
            fpath.unlink()
            self._asset_manager_context._rm(name)
        self._asset_manager_context.sync(check=True)


class AssetManagerContext:
    """
    A subcontext within an :class:`AssetManager`

    Represents a namespace within the :class:`AssetManager`, essentially a
    subdirectory. The context holds assets for a challenge with the same id

    This class is not meant to be constructed directly, use
    :meth:`AssetManager.create_context`
    """

    _asset_manager: "AssetManager"
    _name: str
    _root: Path
    _files: Set[str]
    _files_root: Path
    _manifest_file: Path

    _is_transaction_active: bool

    def __init__(self, asset_manager: "AssetManager", name: str):
        """
        :meta private:
        """
        self._asset_manager = asset_manager
        self._name = name
        self._files = set()
        self._is_transaction_active = False
        self._root = self._asset_manager.root / name
        self._root.mkdir(parents=True, exist_ok=True)
        self._files_root = self._root / "files"
        self._files_root.mkdir(exist_ok=True)
        self._manifest_file = self._root / "manifest.json"

        try:
            with self._manifest_file.open("r") as fd:
                manifest = json.load(fd)
            self._files = set(manifest["files"])
            for fn in list(self._files):
                f = self._get(fn)
                if f.is_symlink() and not f.exists():
                    # Broken symlink; remove it
                    self._files.remove(fn)
                    f.unlink()
            self.sync()
        except FileNotFoundError:
            pass

    def _assert_valid_name(self, name: str) -> None:
        if not _is_valid_name(name):
            raise ValueError(f"Invalid asset name '{name}'")

    def transaction(self) -> AssetManagerTransaction:
        """
        Create a :class:`AssetManagerTransaction`.

        Only one transaction can be created at a time.

        :returns: The transaction
        :raises RuntimeError: when attempting to create a transaction while one already
            exists
        """
        # TODO: better locking mechanism?
        if self._is_transaction_active:
            raise RuntimeError(
                "Attempted to create transaction while one is already in progress"
            )
        self._is_transaction_active = True
        return AssetManagerTransaction(self)

    def sync(self, *, check: bool = True):
        """
        Syncs the manifest for this context to disk

        :param bool check: If true (default), check to make sure all files in the
            manifest exist, and that there are no extra files
        """
        if check:
            disk = set(self._files_root.iterdir())
            files = {self._files_root / f for f in self._files}
            for extra in disk - files:
                warn(
                    RuntimeWarning(
                        f"Unexpected item found in cache: '{str(extra)}'; removing..."
                    )
                )
                if extra.is_dir():
                    shutil.rmtree(extra)
                else:
                    extra.unlink()
            for missing in files - disk:
                raise RuntimeError(f"Cache item missing: '{str(missing)}'")
        with self._manifest_file.open("w") as fd:
            json.dump({"files": sorted(self._files)}, fd)

    def _add(self, name: str, *, force: bool = False) -> None:
        """
        Add an asset to the manifest

        :meta private:
        :param str name: The name of the asset
        :param bool force: If true, do not error if the asset already exists
        """
        self._assert_valid_name(name)
        if not force and name in self._files:
            raise FileExistsError(f"Asset already exists: '{name}'")
        self._files.add(name)

    def _rm(self, name: str, *, force: bool = False) -> None:
        """
        Remove an asset from the manifest

        :meta private:
        :param str name: The name of the asset
        :param bool force: If true, do not error if the asset does not exist
        """
        self._assert_valid_name(name)
        try:
            self._files.remove(name)
        except KeyError:
            if not force:
                raise FileNotFoundError(f"Asset not found: '{name}'")

    def ls(self) -> Iterable[str]:
        """
        List all files within this context

        :returns: The list of asset names
        """
        return self._files

    def _get(self, name: str) -> Path:
        """
        Retrieves the path for an asset with the given name, even if it does not already
        exist

        :meta private:
        """
        self._assert_valid_name(name)
        return self._files_root / name

    def exists(self, name: str) -> bool:
        """
        Queries if an asset exists

        :param str name: The name of the asset
        :returns: Whether or not it exists
        """
        self._assert_valid_name(name)
        return name in self._files

    def get(self, name: str) -> Path:
        """
        Retrieves the asset

        :param str name: The name of the asset
        :returns: The asset
        """
        if not self.exists(name):
            raise FileNotFoundError(f"Asset not found: '{name}'")
        return self._get(name)

    def get_mtime(self, name: str) -> float:
        """
        Retrieves the time an asset was modified

        :param str name: The name of the asset
        :returns: The time the asset was modified (:attr`os.stat_result.st_mtime`)
        """
        return self.get(name).stat().st_mtime

    def clear(self) -> None:
        """
        Clear all files in this context
        """
        for f in self.ls():
            self.get(f).unlink()
        self._files = set()
        self.sync(check=True)


class AssetManager:
    """
    Class for managing assets from challenges that are provided to competitors

    This class manages all assets under a given project.
    """

    project: "Project"
    root: Path

    def __init__(self, project: "rcds.Project"):
        self.project = project
        self.root = self.project.root / ".rcds-cache" / "assets"
        self.root.mkdir(parents=True, exist_ok=True)

    def create_context(self, name: str) -> AssetManagerContext:
        """
        Create a subcontext within the :class:`AssetManager`

        :param str name: The name of the context (challenge id)
        :raises ValueError: if the context name is not valid
        """
        if not _is_valid_name(name):
            raise ValueError(f"Invalid context name '{name}'")
        return AssetManagerContext(self, name)

    def list_context_names(self) -> Iterable[str]:
        """
        List the names of all subcontexts within this :class:`AssetManager`

        :returns: The contexts' names. Call :meth:`create_context` on a name to obtain a
            :class:`AssetManagerContext` object
        """
        for d in self.root.iterdir():
            if not d.is_dir():
                raise RuntimeError(f"Unexpected item found in cache: '{str(d)}'")
            yield d.name



"""
Error types for various rCDS methods
"""

import jsonschema.exceptions  # type: ignore


class ValidationError(ValueError):
    pass


class SchemaValidationError(ValidationError):
    cause: jsonschema.exceptions.ValidationError

    def __init__(self, message: str, cause: jsonschema.exceptions.ValidationError):
        super().__init__(message)

        self.cause = cause



"""
Utility for loading configuration files

Objects only, loading files consisting of arrays is not supported.
"""

import json
from pathlib import Path
from typing import Any, Dict

import yaml


def _normalize_jsonlike(data: Any) -> Dict[str, Any]:
    if data is None:
        return dict()
    return data


def load_yaml(f: Path) -> Dict[str, Any]:
    with f.open("r") as fd:
        return _normalize_jsonlike(yaml.safe_load(fd))


def load_json(f: Path) -> Dict[str, Any]:
    with f.open("r") as fd:
        return _normalize_jsonlike(json.load(fd))


def load_any(f: Path) -> Dict[str, Any]:
    ext = f.suffix[1:]
    if ext in ["yml", "yaml"]:
        return load_yaml(f)
    elif ext in ["json"]:
        return load_json(f)
    else:
        raise Exception("Unsupported extension")


SUPPORTED_EXTENSIONS = ["yml", "yaml", "json"]



from jsonschema import Draft7Validator, validators  # type: ignore

# From
# https://python-jsonschema.readthedocs.io/en/stable/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance # noqa: B950


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        yield from validate_properties(validator, properties, instance, schema)

    return validators.extend(validator_class, {"properties": set_defaults})


DefaultValidatingDraft7Validator = extend_with_default(Draft7Validator)



from base64 import b64encode
from typing import Any, Dict, List, Optional
from urllib.parse import quote, urljoin

import requests
from requests_toolbelt.sessions import BaseUrlSession  # type: ignore


class RCTFAdminV1:

    session: requests.Session

    def __init__(self, endpoint: str, login_token: Optional[str]):
        self.session = BaseUrlSession(urljoin(endpoint, "api/v1/admin/"))

        if login_token is not None:
            login_resp = requests.post(
                urljoin(endpoint, "api/v1/auth/login"), json={"teamToken": login_token}
            ).json()
            if login_resp["kind"] == "goodLogin":
                auth_token = login_resp["data"]["authToken"]
                self.session.headers["Authorization"] = f"Bearer {auth_token}"
            else:
                raise ValueError(
                    f"Invalid login_token provided (reason: {login_resp['kind']})"
                )

    @staticmethod
    def assertResponseKind(response: Any, kind: str) -> None:
        if response["kind"] != kind:
            raise RuntimeError(f"Server error: {response['kind']}")

    def list_challenges(self) -> List[Dict[str, Any]]:
        r = self.session.get("challs").json()
        self.assertResponseKind(r, "goodAdminChallenges")
        return r["data"]

    def put_challenge(self, chall_id: str, data: Dict[str, Any]) -> None:
        r = self.session.put("challs/" + quote(chall_id), json={"data": data}).json()
        self.assertResponseKind(r, "goodChallengeUpdate")

    def delete_challenge(self, chall_id: str) -> None:
        r = self.session.delete("challs/" + quote(chall_id)).json()
        self.assertResponseKind(r, "goodChallengeDelete")

    def create_upload(self, uploads: Dict[str, bytes]) -> Dict[str, str]:
        """
        :param uploads: uploads {name: data}
        :return: urls {name: url}
        """
        if len(uploads) == 0:
            return {}
        payload = [
            {"name": name, "data": "data:;base64," + b64encode(data).decode()}
            for name, data in uploads.items()
        ]
        r = self.session.post("upload", json={"files": payload}).json()
        self.assertResponseKind(r, "goodFilesUpload")
        return {f["name"]: f["url"] for f in r["data"]}

    def get_url_for_files(self, files: Dict[str, str]) -> Dict[str, Optional[str]]:
        """
        :param files: files to get {name: sha256}
        :return: urls {name: url}
        """
        payload = [{"name": name, "sha256": sha256} for name, sha256 in files.items()]
        r = self.session.post("upload/query", json={"uploads": payload}).json()
        self.assertResponseKind(r, "goodUploadsQuery")
        return {f["name"]: f["url"] for f in r["data"]}



import os
from hashlib import sha256
from pathlib import Path
from typing import Any, Dict, Set

import rcds
import rcds.backend
from rcds.util import deep_merge, load_any
from rcds.util.jsonschema import DefaultValidatingDraft7Validator

from .rctf import RCTFAdminV1

options_schema_validator = DefaultValidatingDraft7Validator(
    schema=load_any(Path(__file__).parent / "options.schema.yaml")
)


class ScoreboardBackend(rcds.backend.BackendScoreboard):
    _project: rcds.Project
    _options: Dict[str, Any]
    _adminv1: RCTFAdminV1

    def __init__(self, project: rcds.Project, options: Dict[str, Any]):
        self._project = project
        self._options = options

        for option_key in ["url", "token"]:
            env_key = f"RCDS_RCTF_{option_key.upper()}"
            self._options[option_key] = os.environ.get(
                env_key, self._options.get(option_key, None)
            )

        # FIXME: validate options better
        if not options_schema_validator.is_valid(self._options):
            raise ValueError("Invalid options")

        self._adminv1 = RCTFAdminV1(self._options["url"], self._options["token"])

    def patch_challenge_schema(self, schema: Dict[str, Any]) -> None:
        # Disallow regex flags
        flag_schema = next(
            s for s in schema["properties"]["flag"]["oneOf"] if s["type"] == "object"
        )
        flag_schema["properties"].pop("regex")
        flag_schema["oneOf"] = [
            s for s in flag_schema["oneOf"] if s["required"][0] != "regex"
        ]

        # tiebreakEligible flag
        schema["properties"]["tiebreakEligible"] = {
            "type": "boolean",
            "description": "Whether or not this challenge affects tiebreakers.",
            "default": True,
        }

        # sortWeight
        schema["properties"]["sortWeight"] = {
            "type": "number",
            "description": (
                "A parameter used for ordering when points and solves are equal."
            ),
            "default": 0,
        }

        schema["required"] += ["author", "category", "tiebreakEligible", "sortWeight"]

    def commit(self) -> bool:
        # Validate challenges
        for challenge in self._project.challenges.values():
            self.validate_challenge(challenge)

        for challenge in self._project.challenges.values():
            self.preprocess_challenge(challenge)

        # Begin actual commit
        remote_challenges: Set[str] = set(
            c["id"]
            for c in self._adminv1.list_challenges()
            if c.get("managedBy", None) == "rcds"
        )
        for challenge in self._project.challenges.values():
            if not challenge.config["visible"]:
                continue
            try:
                remote_challenges.remove(challenge.config["id"])
            except KeyError:
                pass
            self.commit_challenge(challenge)
        for chall_id in remote_challenges:
            print(f"Deleting {chall_id}")
            self._adminv1.delete_challenge(chall_id)
        return True

    def validate_challenge(self, challenge: rcds.Challenge) -> None:
        """
        Raises exception on validation fail
        """
        if isinstance(challenge.config["flag"], dict):
            if challenge.config["flag"]["regex"] is not None:
                raise ValueError("rCTF does not support regex flags")
            else:
                raise RuntimeError(
                    'Unexpected content in "flag" key on challenge config'
                )

    def preprocess_challenge(self, challenge: rcds.Challenge) -> None:
        chall_id = challenge.config["id"]
        if "sortOrder" in self._options:
            if chall_id in self._options["sortOrder"]:
                challenge.config["sortWeight"] = -self._options["sortOrder"].index(
                    chall_id
                )

    def commit_challenge(self, challenge: rcds.Challenge) -> None:
        chall_id = challenge.config["id"]
        rctf_challenge: Dict[str, Any] = {"managedBy": "rcds"}
        for common_field in [
            "name",
            "author",
            "category",
            "flag",
            "tiebreakEligible",
            "sortWeight",
        ]:
            rctf_challenge[common_field] = challenge.config[common_field]
        rctf_challenge["description"] = challenge.render_description()
        if "value" in challenge.config:
            # Static value
            rctf_challenge["points"] = {
                "min": challenge.config["value"],
                "max": challenge.config["value"],
            }
        else:
            # No value = dynamically scored
            rctf_challenge["points"] = {
                "min": self._options["scoring"]["minPoints"],
                "max": self._options["scoring"]["maxPoints"],
            }

        am_ctx = challenge.get_asset_manager_context()
        file_hashes: Dict[str, str] = dict()
        for filename in am_ctx.ls():
            h = sha256()
            with am_ctx.get(filename).open("rb") as fd:
                for chunk in iter(lambda: fd.read(5245288), b""):
                    h.update(chunk)
            file_hashes[filename] = h.hexdigest()
        file_urls: Dict[str, str] = {
            f: u
            for f, u in self._adminv1.get_url_for_files(file_hashes).items()
            if u is not None
        }
        deep_merge(
            file_urls,
            self._adminv1.create_upload(
                {
                    name: am_ctx.get(name).read_bytes()
                    for name in am_ctx.ls()
                    if name not in file_urls
                }
            ),
        )
        rctf_challenge["files"] = [
            {"name": name, "url": url} for name, url in file_urls.items()
        ]

        self._adminv1.put_challenge(chall_id, rctf_challenge)


class BackendsInfo(rcds.backend.BackendsInfo):
    HAS_SCOREBOARD = True

    def get_scoreboard(
        self, project: rcds.Project, options: Dict[str, Any]
    ) -> ScoreboardBackend:
        return ScoreboardBackend(project, options)


def get_info() -> BackendsInfo:
    return BackendsInfo()



"""
Functions for loading project config files
"""

from itertools import tee
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Tuple, Union, cast

import jsonschema  # type: ignore

from rcds import errors

from ..util import load_any
from ..util.jsonschema import DefaultValidatingDraft7Validator

config_schema_validator = DefaultValidatingDraft7Validator(
    schema=load_any(Path(__file__).parent / "rcds.schema.yaml"),
    format_checker=jsonschema.draft7_format_checker,
)


def parse_config(
    config_file: Path,
) -> Iterable[Union[errors.ValidationError, Dict[str, Any]]]:
    """
    Load and validate a config file, returning both the config and any
    errors encountered.

    :param pathlib.Path config_file: The challenge config to load
    :returns: Iterable containing any errors (all instances of
        :class:`rcds.errors.ValidationError`) and the parsed config. The config will
        always be last.
    """
    # root = config_file.parent
    config = load_any(config_file)
    schema_errors: Iterable[errors.SchemaValidationError] = (
        errors.SchemaValidationError(str(e), e)
        for e in config_schema_validator.iter_errors(config)
    )
    # Make a duplicate to check whethere there are errors returned
    schema_errors, schema_errors_dup = tee(schema_errors)
    # This is the same test as used in Validator.is_valid
    if next(schema_errors_dup, None) is not None:
        yield from schema_errors
    yield config


def check_config(
    config_file: Path,
) -> Tuple[Optional[Dict[str, Any]], Optional[Iterable[errors.ValidationError]]]:
    """
    Load and validate a config file, returning any errors encountered.

    If the config file is valid, the tuple returned contains the loaded config as
    the first element, and the second element is None. Otherwise, the second
    element is an iterable of errors that occurred during validation

    This method wraps :func:`parse_config`.

    :param pathlib.Path config_file: The challenge config to load
    """
    load_data = parse_config(config_file)
    load_data, load_data_dup = tee(load_data)
    first = next(load_data_dup)
    if isinstance(first, errors.ValidationError):
        validation_errors = cast(
            Iterable[errors.ValidationError],
            filter(lambda v: isinstance(v, errors.ValidationError), load_data),
        )
        return (None, validation_errors)
    else:
        return (first, None)


def load_config(config_file: Path) -> Dict[str, Any]:
    """
    Loads a config file, or throw an exception if it is not valid

    This method wraps :func:`check_config`, and throws the first error returned
    if there are any errors.

    :param pathlib.Path config_file: The challenge config to load
    :returns: The loaded config
    """
    config, errors = check_config(config_file)
    if errors is not None:
        raise next(iter(errors))
    # errors is None
    assert config is not None
    return config



from pathlib import Path
from typing import Any, Dict, Optional

import docker  # type: ignore
from jinja2 import Environment

from rcds.util import SUPPORTED_EXTENSIONS, find_files

from ..backend import BackendContainerRuntime, BackendScoreboard, load_backend_module
from ..challenge import Challenge, ChallengeLoader
from . import config
from .assets import AssetManager


class Project:
    """
    An rCDS project; the context that all actions are done within
    """

    root: Path
    config: dict
    challenges: Dict[Path, Challenge]
    challenge_loader: ChallengeLoader

    asset_manager: AssetManager

    container_backend: Optional[BackendContainerRuntime] = None
    scoreboard_backend: Optional[BackendScoreboard] = None

    jinja_env: Environment
    docker_client: Any

    def __init__(
        self, root: Path, docker_client: Optional[docker.client.DockerClient] = None
    ):
        """
        Create a project
        """
        root = root.resolve()
        try:
            cfg_file = find_files(
                ["rcds"], SUPPORTED_EXTENSIONS, path=root, recurse=False
            )["rcds"]
        except KeyError:
            raise ValueError(f"No config file found at '{root}'")
        self.root = root
        self.config = config.load_config(cfg_file)
        self.challenge_loader = ChallengeLoader(self)
        self.challenges = dict()
        self.asset_manager = AssetManager(self)
        self.jinja_env = Environment(autoescape=False)
        if docker_client is not None:
            self.docker_client = docker_client
        else:
            self.docker_client = docker.from_env()

    def load_all_challenges(self) -> None:
        for ext in SUPPORTED_EXTENSIONS:
            for chall_file in self.root.rglob(f"challenge.{ext}"):
                path = chall_file.parent
                print(f"Loading: {path}")
                self.challenges[
                    path.relative_to(self.root)
                ] = self.challenge_loader.load(path)

    def get_challenge(self, relPath: Path) -> Challenge:
        return self.challenges[relPath]

    def load_backends(self) -> None:
        for backend_config in self.config["backends"]:
            backend_info = load_backend_module(backend_config["resolve"])
            if self.scoreboard_backend is None and backend_info.HAS_SCOREBOARD:
                self.scoreboard_backend = backend_info.get_scoreboard(
                    self, backend_config["options"]
                )
            if self.container_backend is None and backend_info.HAS_CONTAINER_RUNTIME:
                self.container_backend = backend_info.get_container_runtime(
                    self, backend_config["options"]
                )
        # TODO: maybe don't reinitialize here?
        self.challenge_loader = ChallengeLoader(self)



from abc import ABC, abstractmethod
from importlib import import_module
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    import rcds


class BackendBase(ABC):
    def patch_challenge_schema(self, schema: Dict[str, Any]):
        pass


class BackendScoreboard(BackendBase):
    @abstractmethod
    def commit(self) -> bool:
        raise NotImplementedError()


class BackendContainerRuntime(BackendBase):
    @abstractmethod
    def commit(self) -> bool:
        raise NotImplementedError()


class BackendsInfo:
    HAS_SCOREBOARD: bool = False
    HAS_CONTAINER_RUNTIME: bool = False

    def get_scoreboard(
        self, project: "rcds.Project", options: Dict[str, Any]
    ) -> BackendScoreboard:
        raise NotImplementedError()

    def get_container_runtime(
        self, project: "rcds.Project", options: Dict[str, Any]
    ) -> BackendContainerRuntime:
        raise NotImplementedError()


def load_backend_module(name: str) -> BackendsInfo:
    try:
        module = import_module(f"rcds.backends.{name}")
    except ModuleNotFoundError:
        module = import_module(name)
    return module.get_info()  # type: ignore



import re
from copy import deepcopy
from itertools import tee
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Optional,
    Pattern,
    Tuple,
    Union,
    cast,
)
from warnings import warn

import jsonschema  # type: ignore

from rcds import errors

from ..util import deep_merge, load_any
from ..util.jsonschema import DefaultValidatingDraft7Validator

if TYPE_CHECKING:
    from rcds import Project


config_schema = load_any(Path(__file__).parent / "challenge.schema.yaml")


class TargetNotFoundError(errors.ValidationError):
    pass


class TargetFileNotFoundError(TargetNotFoundError):
    target: Path

    def __init__(self, message: str, target: Path):
        super().__init__(message)
        self.target = target


class InvalidFlagError(errors.ValidationError):
    pass


class ConfigLoader:
    """
    Object that manages loading challenge config files
    """

    project: "Project"
    config_schema: Dict[str, Any]
    config_schema_validator: Any
    _flag_regex: Optional[Pattern[str]] = None

    def __init__(self, project: "Project"):
        """
        :param rcds.Project project: project context to use
        """
        self.project = project
        self.config_schema = deepcopy(config_schema)

        # Load flag regex if present
        if "flagFormat" in self.project.config:
            self._flag_regex = re.compile(f"^{self.project.config['flagFormat']}$")

        # Backend config patching
        for backend in [
            self.project.container_backend,
            self.project.scoreboard_backend,
        ]:
            if backend is not None:
                backend.patch_challenge_schema(self.config_schema)
        self.config_schema_validator = DefaultValidatingDraft7Validator(
            schema=self.config_schema, format_checker=jsonschema.draft7_format_checker
        )

    def _apply_defaults(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Apply project-level defaults
        """
        try:
            root_defaults = deepcopy(self.project.config["defaults"])
        except KeyError:
            # No defaults present
            return config

        container_defaults = root_defaults.pop("containers", None)
        expose_defaults = root_defaults.pop("expose", None)

        # Array types with no sensible defaults - ignore them
        root_defaults.pop("provide", None)

        config = deep_merge(root_defaults, config)
        if container_defaults is not None and "containers" in config:
            for container_name, container_config in config["containers"].items():
                config["containers"][container_name] = deep_merge(
                    dict(), container_defaults, container_config
                )
        if expose_defaults is not None and "expose" in config:
            for expose_config in config["expose"].values():
                for i, expose_port in enumerate(expose_config):
                    expose_config[i] = deep_merge(dict(), expose_defaults, expose_port)
        return config

    def parse_config(
        self, config_file: Path
    ) -> Iterable[Union[errors.ValidationError, Dict[str, Any]]]:
        """
        Load and validate a config file, returning both the config and any
        errors encountered.

        :param pathlib.Path config_file: The challenge config to load
        :returns: Iterable containing any errors (all instances of
            :class:`rcds.errors.ValidationError`) and the parsed config. The config will
            always be last.
        """
        root = config_file.parent
        relative_path = root.resolve().relative_to(self.project.root.resolve())
        config = load_any(config_file)

        config.setdefault("id", root.name)  # derive id from parent directory name

        config = self._apply_defaults(config)

        if len(relative_path.parts) >= 2:
            # default category name is the parent of the challenge directory
            config.setdefault("category", relative_path.parts[-2])

        schema_errors: Iterable[errors.SchemaValidationError] = (
            errors.SchemaValidationError(str(e), e)
            for e in self.config_schema_validator.iter_errors(config)
        )
        # Make a duplicate to check whethere there are errors returned
        schema_errors, schema_errors_dup = tee(schema_errors)
        # This is the same test as used in Validator.is_valid
        if next(schema_errors_dup, None) is not None:
            yield from schema_errors
        else:
            if "expose" in config:
                if "containers" not in config:
                    yield TargetNotFoundError(
                        "Cannot expose ports without containers defined"
                    )
                else:
                    for key, expose_objs in config["expose"].items():
                        if key not in config["containers"]:
                            yield TargetNotFoundError(
                                f'`expose` references container "{key}" but '
                                f"it is not defined in `containers`"
                            )
                        else:
                            for expose_obj in expose_objs:
                                if (
                                    expose_obj["target"]
                                    not in config["containers"][key]["ports"]
                                ):
                                    yield TargetNotFoundError(
                                        f"`expose` references port "
                                        f'{expose_obj["target"]} on container '
                                        f'"{key}" which is not defined'
                                    )
            if "provide" in config:
                for f in config["provide"]:
                    if isinstance(f, str):
                        f = Path(f)
                    else:
                        f = Path(f["file"])
                    if not (root / f).is_file():
                        yield TargetFileNotFoundError(
                            f'`provide` references file "{str(f)}" which does not '
                            f"exist",
                            f,
                        )
            if "flag" in config:
                if isinstance(config["flag"], dict):
                    if "file" in config["flag"]:
                        f = Path(config["flag"]["file"])
                        f_resolved = root / f
                        if f_resolved.is_file():
                            with f_resolved.open("r") as fd:
                                flag = fd.read().strip()
                            config["flag"] = flag
                        else:
                            yield TargetFileNotFoundError(
                                f'`flag.file` references file "{str(f)}" which does '
                                f"not exist",
                                f,
                            )
                if isinstance(config["flag"], str):
                    if self._flag_regex is not None and not self._flag_regex.match(
                        config["flag"]
                    ):
                        yield InvalidFlagError(
                            f'Flag "{config["flag"]}" does not match the flag format'
                        )
                    if config["flag"].count("\n") > 0:
                        warn(
                            RuntimeWarning(
                                "Flag contains multiple lines; is this intended?"
                            )
                        )
        yield config

    def check_config(
        self, config_file: Path
    ) -> Tuple[Optional[Dict[str, Any]], Optional[Iterable[errors.ValidationError]]]:
        """
        Load and validate a config file, returning any errors encountered.

        If the config file is valid, the tuple returned contains the loaded config as
        the first element, and the second element is None. Otherwise, the second
        element is an iterable of errors that occurred during validation

        This method wraps :meth:`parse_config`.

        :param pathlib.Path config_file: The challenge config to load
        """
        load_data = self.parse_config(config_file)
        load_data, load_data_dup = tee(load_data)
        first = next(load_data_dup)
        if isinstance(first, errors.ValidationError):
            validation_errors = cast(
                Iterable[errors.ValidationError],
                filter(lambda v: isinstance(v, errors.ValidationError), load_data),
            )
            return (None, validation_errors)
        else:
            return (first, None)

    def load_config(self, config_file: Path) -> Dict[str, Any]:
        """
        Loads a config file, or throw an exception if it is not valid

        This method wraps :meth:`check_config`, and throws the first error returned
        if there are any errors.

        :param pathlib.Path config_file: The challenge config to load
        :returns: The loaded config
        """
        config, errors = self.check_config(config_file)
        if errors is not None:
            raise next(iter(errors))
        # errors is None
        assert config is not None
        return config



from itertools import chain
from pathlib import Path
from typing import Dict, List, Optional

from .load import SUPPORTED_EXTENSIONS


def find_files(
    names: List[str],
    extensions: List[str],
    path: Optional[Path] = None,
    recurse: bool = True,
) -> Dict[str, Path]:
    if path is None:
        path = Path.cwd().resolve()
    foundNames = set(names)
    found = dict()
    dirList = chain([path], path.parents) if recurse else [path]
    for d in dirList:
        for f in filter(lambda f: f.is_file(), d.iterdir()):
            if f.suffix[1:] in extensions and f.stem in foundNames:
                found[f.stem] = f
                foundNames.remove(f.stem)
    return found


def find_cfgs(path: Optional[Path] = None) -> Dict[str, Path]:
    return find_files(["rcds", "challenge"], SUPPORTED_EXTENSIONS, path)



from .backend import BackendBase  # noqa: F401
from .backend import BackendContainerRuntime  # noqa: F401
from .backend import BackendScoreboard  # noqa: F401
from .backend import BackendsInfo  # noqa: F401
from .backend import load_backend_module  # noqa: F401



from copy import deepcopy
from typing import List, Tuple, Union


def deep_merge(a: dict, *rest: dict) -> dict:
    """
    Deep merge remaining arguments from left to right into `a`. Mutates `a`,
    does not mutate any other arguments.

    :param a dict:
    :return: the merged dictionary
    :rtype: dict
    """
    _rest: Union[List[dict], Tuple[dict, ...]] = rest
    while len(_rest) > 0:
        b, *_rest = _rest
        for key in b:
            if isinstance(b[key], dict):
                if key in a and isinstance(a[key], dict):
                    deep_merge(a[key], b[key])
                else:
                    a[key] = deepcopy(b[key])
            else:
                a[key] = b[key]
    return a
